//testperm.c  ## benes extracted 
// (c) 2011..2014 by Jasper L. Neumann
// www.sirrida.de / programming.sirrida.de
// E-Mail: info@sirrida.de
// Granted to the public domain
// First version: 2011-02
// Last change: 2014-09-16
//#include "general.c"
// Choose a working size by including the needed perm_b*.c here:
//#include "perm_b32.h" 
//perm_b64.h
//#include "perm_bas.c"  
#include <stdlib.h>
#include <stdio.h>
#define ld_bits 6
//typedef t_64u t_bits;  // 1<<ld_bits bits, unsigned perm_b64.h
typedef unsigned long long  t_bits; // 64 bit
typedef int t_int;  // int32
typedef unsigned int t_uint;  // int32
#define bits (1 << ld_bits)         // 1<<6 == 64
typedef t_int t_bit_index;  // subrange 0..bits-1; -1:don't care
typedef t_bit_index ta_index[bits]; // perm_bxx.h
typedef struct {
  // This structure is used to hold the configuration of
  // butterfly-based operations as well as compress and expand.
  t_bits cfg[ld_bits];  // butterfly configuration
  t_bits mask;  // saved mask, for compress/expand (unused????)
  } tr_bfly;
typedef struct {
  tr_bfly b1,b2;
 } tr_benes;
typedef t_uint    t_subword; 
typedef t_subword ta_subword[ld_bits]; //perm_bxx.h
const ta_subword a_stage_bwd = {5,4,3,2,1,0}; // perm_b64.h
#define no_index ((t_bit_index)(-1))
#define lo_bit (t_bits)(1)
void invert_perm(const ta_index src, ta_index tgt) {
  t_int i;
  for (i = 0; i <= bits-1; ++i) tgt[i] = no_index; 
  for (i = 0; i <= bits-1; ++i) if (src[i] != no_index) tgt[src[i]] = i;
}
static void exchange_bit_index(t_bit_index* a, t_bit_index* b) { // INLINE
  t_bit_index q; q = *a; *a = *b; *b = q;
}
void gen_benes_ex(tr_benes* self, const ta_index c_tgt,
   const ta_subword a_stage) {
// Generate a configuration for the Benes network with variable stage order.
// Use benes_fwd_ex and benes_bwd_ex.
// Algorithm as sketched by Donal E. Knuth,
//   The art of computer programming, vol. 4, pre-fascicle 1a
// Implemented 2011-12-02 by Jasper L. Neumann
// Modified 2012-08-31 to allow for "don't care" entries.
  ta_index src, inv_src;
  ta_index tgt, inv_tgt;
  t_int stage;
  t_int mask; 
  t_bits cfg_src,cfg_tgt;
  t_bits src_set;
  t_int main_idx, aux_idx, src_idx, tgt_idx, idx2;
  t_int stage_idx;
  t_int s;
  for (s = 0; s <= bits-1; ++s) { src[s] = no_index; tgt[s] = no_index; }
  for (s = 0; s <= bits-1; ++s) {
    if (c_tgt[s] != no_index) { tgt[s] = s; src[c_tgt[s]] = s; }
  }
  invert_perm(src,inv_src);
  invert_perm(tgt,inv_tgt);
  for (stage_idx = 0; stage_idx <= ld_bits-1; ++stage_idx) {
    stage = a_stage[stage_idx];   src_set = 0;
    mask = lo_bit << stage;       cfg_src = 0;    cfg_tgt = 0;
    for (main_idx = 0; main_idx <= bits-1; ++main_idx) // This order to meet
    if ((main_idx & mask) == 0)   // low only
    for (aux_idx = 0; aux_idx <= 1; ++aux_idx) {
      src_idx = main_idx+(aux_idx << stage);
      if (((lo_bit << src_idx) & src_set) == 0)   // yet unhandled
      if (src[src_idx] != no_index)  // not open
      do {
        src_set = src_set | (lo_bit << src_idx);
        tgt_idx = inv_tgt[src[src_idx]];
        if (tgt[tgt_idx] == no_index) break;  // open end

        if (((src_idx ^ tgt_idx) & mask) == 0) // straight
          tgt_idx = tgt_idx ^ mask;
        else {
          // cross
          cfg_tgt = cfg_tgt | (lo_bit << (tgt_idx & ~mask));
          idx2 = tgt_idx ^ mask;
          exchange_bit_index(&tgt[tgt_idx], &tgt[idx2]);
          inv_tgt[tgt[idx2]] = idx2;
          if (tgt[tgt_idx] != no_index) inv_tgt[tgt[tgt_idx]] = tgt_idx;
          }  
        if (tgt[tgt_idx] == no_index) break;  // open end
        src_idx = inv_src[tgt[tgt_idx]];
        if (((src_idx ^ tgt_idx) & mask) == 0) { // straight
          src_set = src_set | (lo_bit << src_idx);
          src_idx = src_idx ^ mask;
        } else { // cross
          cfg_src = cfg_src | (lo_bit << (src_idx & ~mask));
          idx2 = src_idx ^ mask;
          src_set = src_set | (lo_bit << idx2);
          exchange_bit_index(&src[src_idx], &src[idx2]);
          inv_src[src[idx2]] = idx2;
          if (src[src_idx] != no_index) inv_src[src[src_idx]] = src_idx;
          }  
        if (src[src_idx] == no_index) break;  // open end
        if (((lo_bit << src_idx) & src_set) != 0) break;
        } while (1);
      } // if      
    self->b1.cfg[stage] = cfg_src;
    self->b2.cfg[stage] = cfg_tgt;
    }
  // Reduce inner stages to one (not needed)
  // self->b2.cfg[0] = self->b2.cfg[0] ^ self->b1.cfg[0];
  // self->b1.cfg[0] = 0;
}
void gen_benes(tr_benes* self, const ta_index c_tgt) {
  gen_benes_ex(self,c_tgt,a_stage_bwd);  // standard Benes order
}
void random_perm(ta_index tgt) {
  t_int j,q;
  t_int x;
  for (q = 0; q <= bits-1; ++q) tgt[q] = q;
  for (q = 1; q <= bits-1; ++q) {
    j = random() % (q+1);  // 0..q
    x = tgt[q]; tgt[q] = tgt[j]; tgt[j] = x;
  }
}
int main(int an, char *av[]) {
  int i, j, k, a1;
  unsigned long long mtab[6]={
    0x5555555555555555L,
    0x3333333333333333L,
    0x0f0f0f0f0f0f0f0fL,
    0x00ff00ff00ff00ffL,
    0x0000ffff0000ffffL,
    0x00000000ffffffffL
    };
  ta_index p1, p2;          // int[64]
  ta_index p3={0,7,13,19,20,27,33,39,6,12,18,26,32,38,5,11,17,25,31,37,4,3,10,16,24,23,30,36,2,9,15,
22,29,35,1,8,14,21,28,34,40,41,42,43,44,45,46,47,48,49,50,51,52,53,54,55,56,57,58,59,60,61,62,63};
  tr_benes benes;  // perm_bas.h
  if (an>1) a1=atoi(av[1]); 
  for(i=0;i<64;i++)p1[i]=p3[i];
  if(a1==1){j=p1[ 0]; p1[ 0]=p1[ 3]; p1[ 3]=p1[ 4]; p1[ 4]=j;}
  if(a1==2){j=p1[ 1]; p1[ 1]=p1[ 19]; p1[ 19]=j;}
  if(a1==3)for(i=0;i<64;i++)p1[i]=(i+4)%64; // shift
  if(a1==4)for(i=0;i<64;i++)p1[i]=(i+1)%64; // shift
  if(a1==5)for(i=0;i<64;i++)p1[i]=(i+63)%64; // shift
  if(a1==6)for(i=0;i<64;i++)p1[i]=(i+8)%64; // shift
  if(a1==7)for(i=0;i<64;i++)p1[i]=i^1; // shift
  if(a1==8)for(i=0;i<64;i++)p1[i]=63-i; // inversion
  if(a1>=9){srand(a1);random_perm(p1);} // perm_bas.c p1[i=0..63]=i; + rnd_swaps
  for(i=0;i<64;i++)printf(" %d", (int)p1[i]);printf(" =perm\n");
  for(i=0;i<64;i++)p2[i]=p1[i]; // copy to compare later
  gen_benes(&benes,p1); // perm_bas.c
//  for(i=0;i<6;i++)printf(" %016Lx =benes.b1[%d]\n",benes.b1.cfg[i],i);
//  for(i=0;i<6;i++)printf(" %016Lx =benes.b2[%d]\n",benes.b2.cfg[i],i);
 // proof:
  for(i=0;i<64;i++)p1[i]=i;
  for(i=5;i>=0;i--) {  // BFLY
    for (j=0;j<64;j++)
     if ((mtab[i]>>j&1) && (benes.b1.cfg[i]>>j&1)) {
//        printf(" exchange %d %d\n",j,j+(1<<i));
        k=p1[j];
          p1[j]=p1[j+(1<<i)];
                   p1[j+(1<<i)]=k;
        
      }
      printf(" %016Lx =benes.b1[%d].%d\n",benes.b1.cfg[i],i,1<<i);
//      for(j=0;j<64;j++)printf(" %2d", (int)p1[j]);printf(" =perm\n");

  }
  for(i=0;i<6;i++) {  // iBFLY
    for (j=0;j<64;j++)
     if ((mtab[i]>>j&1) && (benes.b2.cfg[i]>>j&1)) {
//        printf(" exchange %d %d\n",j,j+(1<<i));
        k=p1[j];
          p1[j]=p1[j+(1<<i)];
                   p1[j+(1<<i)]=k;
        
      }
      printf(" %016Lx =benes.b2[%d].%d\n",benes.b2.cfg[i],i,1<<i);
//      for(j=0;j<64;j++)printf(" %2d", (int)p1[j]);printf(" =perm\n");

  }
  for(i=0;i<64;i++)printf(" %d", (int)p1[i]);printf(" =perm\n");
  for(i=0;i<64 && p2[i]==p1[i];i++); // compare
  if (i==64) printf("OK\n");
  else printf("ERR at pos %d\n",i);
  return 0;
}

/*

2-8
 0 1 8 3 4 5 6 7 2 9 10 11 12 13 14 15 16 17 18 19 20 21 22 23 24 25 26 27 28 29
 30 31 32 33 34 35 36 37 38 39 40 41 42 43 44 45 46 47 48 49 50 51 52 53 54 55 5
6 57 58 59 60 61 62 63 =perm
 0000000000000000 =benes.b1[5].32
 0000000000000000 =benes.b1[4].16
 0000000000000004 =benes.b1[3].8
 0000000000000000 =benes.b1[2].4
 0000000000000000 =benes.b1[1].2
 0000000000000000 =benes.b1[0].1
 0000000000000000 =benes.b2[0].1
 0000000000000100 =benes.b2[1].2
 0000000000000000 =benes.b2[2].4
 0000000000000004 =benes.b2[3].8
 0000000000000000 =benes.b2[4].16
 0000000000000000 =benes.b2[5].32

2-9
 0 1 9 3 4 5 6 7 8 2 10 11 12 13 14 15 16 17 18 19 20 21 22 23 24 25 26 27 28 29
 30 31 32 33 34 35 36 37 38 39 40 41 42 43 44 45 46 47 48 49 50 51 52 53 54 55 5
6 57 58 59 60 61 62 63 =perm
 0000000000000000 =benes.b1[5].32
 0000000000000000 =benes.b1[4].16
 0000000000000004 =benes.b1[3].8
 0000000000000000 =benes.b1[2].4
 0000000000000200 =benes.b1[1].2
 0000000000000000 =benes.b1[0].1
 0000000000000400 =benes.b2[0].1
 0000000000000200 =benes.b2[1].2
 0000000000000000 =benes.b2[2].4
 0000000000000004 =benes.b2[3].8
 0000000000000000 =benes.b2[4].16
 0000000000000000 =benes.b2[5].32
 0 1 9 3 4 5 6 7 8 2 10 11 12 13 14 15 16 17 18 19 20 21 22 23 24 25 26 27 28 29
 30 31 32 33 34 35 36 37 38 39 40 41 42 43 44 45 46 47 48 49 50 51 52 53 54 55 5
6 57 58 59 60 61 62 63 =perm
OK

1-19
 0 19 2 3 4 5 6 7 8 9 10 11 12 13 14 15 16 17 18 1 20 21 22 23 24 25 26 27 28 29
 30 31 32 33 34 35 36 37 38 39 40 41 42 43 44 45 46 47 48 49 50 51 52 53 54 55 5
6 57 58 59 60 61 62 63 =perm
 0000000000000000 =benes.b1[5].32
 exchange 3 19
 0000000000000008 =benes.b1[4].16
 0000000000000000 =benes.b1[3].8
 0000000000000000 =benes.b1[2].4
 0000000000000000 =benes.b1[1].2
 0000000000000000 =benes.b1[0].1
 0000000000000000 =benes.b2[0].1
 exchange 1 3
 0000000000000002 =benes.b2[1].2
 0000000000000000 =benes.b2[2].4
 0000000000000000 =benes.b2[3].8
 exchange 3 19
 0000000000000008 =benes.b2[4].16
 0000000000000000 =benes.b2[5].32
 0 19 2 3 4 5 6 7 8 9 10 11 12 13 14 15 16 17 18 1 20 21 22 23 24 25 26 27 28 29
 30 31 32 33 34 35 36 37 38 39 40 41 42 43 44 45 46 47 48 49 50 51 52 53 54 55 5
6 57 58 59 60 61 62 63 =perm
OK



*/
